@@ -4,6 +4,7 @@ This module contans Bubbleio class to handle Bubble.io API.
 https://manual.bubble.io/core-resources/api/data-api
 """
 
+from os import initgroups
 import requests
 import logging
 import pandas as pd
@@ -179,7 +180,13 @@ class Bubbleio:
             "results"
         ]
 
-    def get_all_results(self, typename, constraints=None):
+    def get_all_results(
+        self,
+        typename,
+        constraints=None,
+        progress_callback=None,
+        progress_callback_resolution=0.1,
+    ):
         """Get all intems of one "things" type. The function iterates with limit and cursor
         to get all items
 
@@ -189,25 +196,49 @@ class Bubbleio:
                                 This parameter should be an list of constraints, e.g., objects with a ``key``,
                                 ``constraint_type``, and most of the time a ``value``.
                                 See :meth:`~bubbleio.bubbleio.Bubbleio.get` example.
+            progress_callback: Function to be called to communicate progress
+            progress_callback_resolution: Parameter user to tune the frequency of callback_progress function.
 
         Returns:
             List: The list of all items of the type.
         """
         response = self.get(typename, constraints=constraints)
+        cursor = 100  # Because default limit value is 100
 
         remaining = response["remaining"]
         records = response["results"]
 
-        cursor = 100  # Because default limit value is 100
+        item_count = len(response["results"]) + response["remaining"]
+        item_processed = len(response["results"])
+        if progress_callback:
+            progress = item_processed / item_count
+            progress_callback(progress)
+            progress_buffer = progress  # to handle resolution
+
+        self.logger.info("%s items in total. Processing..." % (item_count))
 
         while remaining > 0:
-            self.logger.info(
-                "Querying table %s,  : %s items remaining" % (typename, remaining)
-            )
             response = self.get(typename, cursor=cursor, constraints=constraints)
             records.extend(response["results"])
+            item_processed = cursor + len(response["results"])
+
+            if progress_callback:
+                progress = item_processed / item_count
+                # if progress greater than resolution --> callback
+                if (
+                    (progress - progress_buffer) > progress_callback_resolution
+                ) or progress == 1:
+                    progress_callback(progress)
+                    self.logger.info(
+                        "Progress %s / %s. Callback sent" % (item_processed, item_count)
+                    )
+                    progress_buffer = progress
+
             cursor = cursor + 100
             remaining = response["remaining"]
+            self.logger.debug(
+                "Querying table %s  : %s items remaining" % (typename, remaining)
+            )
 
         return records
 
@@ -223,6 +254,8 @@ class Bubbleio:
                                 This parameter should be an list of constraints, e.g., objects with a ``key``,
                                 ``constraint_type``, and most of the time a ``value``.
                                 See :meth:`~bubbleio.bubbleio.Bubbleio.get` example.
+            progress_callback: Function to be called to communicate progress
+            progress_callback_resolution: Parameter user to tune the frequency of callback_progress function.
 
         Returns:
             Pandas.DataFrame: The list of all items of the type.
@@ -243,7 +276,14 @@ class Bubbleio:
         )
         return df
 
-    def get_all_results_as_df(self, typename, joins=None, constraints=None):
+    def get_all_results_as_df(
+        self,
+        typename,
+        joins=None,
+        constraints=None,
+        progress_callback=None,
+        progress_callback_resolution=0.1,
+    ):
         """Returns all results as a Pandas.DataFrame
 
         Args:
@@ -284,7 +324,14 @@ class Bubbleio:
             idFoo1  value       value       idBar1  idBar1          value              value
             idFoo2  value       value       idBar2  idBar2          value              value
         """
-        df = pd.DataFrame(self.get_all_results(typename, constraints=constraints))
+        df = pd.DataFrame(
+            self.get_all_results(
+                typename,
+                constraints=constraints,
+                progress_callback=progress_callback,
+                progress_callback_resolution=progress_callback_resolution,
+            )
+        )
         if joins:
             for j_param in joins:
                 foreign_table = self.get_all_results_as_df(j_param["typename"])
@@ -301,3 +348,31 @@ class Bubbleio:
                 except KeyError as e:
                     self.logger.warning("Join impossible (KeyError): %s" % (e))
         return df
+
+    def patch(self, typename, unique_id, values, ignore_errors=False):
+        """Python implementaiton of PATCH API call
+
+        To modify a thing, send a PATCH request to the same endpoint you use to retrieve it.
+        The body of the request should be a JSON object with a list of keys and values to modify.
+
+        Args:
+            typename (str): The type of "things" you are querying...
+            unique_id (str): The unique ID of the thing
+            values (dict): Dictionary updated values
+
+        Returns:
+            None
+        """
+
+        self.logger.info("PATCH call on type %s." % (typename))
+
+        r = requests.patch(
+            self.api_root + "/" + typename + "/" + unique_id,
+            headers=self.headers(),
+            json=values,
+        )
+        if r.status_code != 204:
+            if ignore_errors:
+                self.logger.warning("PATCH ERROR %s: %s" % (r.status_code, r.reason))
+            else:
+                r.raise_for_status()
